/*-
 * Copyright (C)2014 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2014 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n\
@(#)V.M "VERS;
static const char rcsid[] = "$Id: common.c,v 1.1 2014/08/14 09:34:37 babolo Exp $";
#endif /* not lint */

#define BLIN_COMPAT  3

#include <sys/types.h>
#include <sysexits.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#include <string.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include "auwith_int.h"

char *
/*****************************************************************************************************
 *****************************************************************************************************/
auwith_todev(const char *o, char *d) {                                                           /****
 *****************************************************************************************************
 *****************************************************************************************************/
    char *s = NULL;
    int c;

    if  (o[0] == '/') {
        s = strdup(o);
    } else if ((o[0] >= '0') && (o[0] <= '9')) {
        errno = 0;
        c = strtoul(o, (char**)&o, 0);
        if  (errno) {
            ifBLIN_QW0("-a strtoul");
            s = NULL;
            goto out;
        } else if (*o) {
            ifBLIN_QX0("-a text");
            errno = EPROTO;
            s = NULL;
            goto out;
        }
        if  (0 > asprintf(&s, "%s%u", d, c)) {
            ifBLIN_QW0("-a asprintf N");
            s = NULL;
            goto out;
        }
    } else {
        if  (0 > asprintf(&s, "/dev/%s", o)) {
            ifBLIN_QW0("-a asprintf /dev/");
            s = NULL;
            goto out;
    }   }
out:
    return(s);
}

int
/*****************************************************************************************************
 *****************************************************************************************************/
auwith_totime(const char *o, struct timespec *tr) {                                              /****
 *****************************************************************************************************
 *****************************************************************************************************/
    int ex = EX_OK;

    errno = 0;
    tr->tv_nsec = 0;
    tr->tv_sec = strtoul(o, (char**)&o, 10);
    if  (errno) {
        ifBLIN_QW0("-t strtoul");
        EXOUT(EX_USAGE);
    } else if (*o == '.') {
        int l;

        l = strlen(++o);
        errno = 0;
        tr->tv_nsec = strtoul(o, (char**)&o, 10);
        if  (errno) {
            ifBLIN_QW0("-t .strtoul");
            EXOUT(EX_USAGE);
        } else if (*o) {
            ifBLIN_QX0("-t .text");
            ERROUT(EX_USAGE, EPROTO);
        }
        for (; l < 9; ++l) tr->tv_nsec *= 10;
        for (; l > 9; --l) tr->tv_nsec /= 10;
    } else if (*o) {
        ifBLIN_QX0("-t text");
        ERROUT(EX_USAGE, EPROTO);
    }
out:
    return(ex);
}

auwith_cell *
/*****************************************************************************************************
 *****************************************************************************************************/
auwith_cella(auwith_pid *pidm, u_int i, u_int o) {                                                /****
 *****************************************************************************************************
 *****************************************************************************************************/
    auwith_cell *c = NULL;

    if  (i >= pidm->ninp) {
        ifBLIN_QX1("INP %u >= %u", i, pidm->ninp);
        goto out;
    }
    if  (o >= pidm->nout) {
        ifBLIN_QX1("OUT %u >= %u", o, pidm->nout);
        goto out;
    }
    c = &pidm->matrix[pidm->nout * i + o];
out:
    return(c);
}

int
/*****************************************************************************************************
 *****************************************************************************************************/
auwith_sfafer(auwith_sfaf *sfaf) {                                                               /****
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (cell->flags & BLIN_MASK)
    auwith_time period, t, meanw, meanr;
    int faser;

    if  (!sfaf->tick[0] || !sfaf->tick[1] || !sfaf->tick[2] || !sfaf->tick[3]) {
        faser = 0;
    } else {
        if  (sfaf->tick[0] > sfaf->tick[2]) {
            t = sfaf->tick[0];
            sfaf->tick[0] = sfaf->tick[2];
            sfaf->tick[2] = t;
        }
        if  (sfaf->tick[1] > sfaf->tick[3]) {
            t = sfaf->tick[1];
            sfaf->tick[1] = sfaf->tick[3];
            sfaf->tick[3] = t;
        }
        period = sfaf->tick[2] - sfaf->tick[0] + sfaf->tick[3] - sfaf->tick[1];
        meanw = sfaf->tick[0] + sfaf->tick[2];
        meanr = sfaf->tick[1] + sfaf->tick[3];
        if  (meanw > meanr) {
            faser = (int)(sfaf->chainlen * AUWITH_FASMUL)
                  - (int)(((meanw - meanr) * AUWITH_FASMUL / period) % AUWITH_FASMUL)
            ;
        } else if ((!!sfaf->chainlen) || ((2 * period + meanw) > meanr)) {
            faser = (((meanr - meanw) * AUWITH_FASMUL / period) % AUWITH_FASMUL)
                  + (sfaf->chainlen * AUWITH_FASMUL)
            ;
        } else {
            sfaf->tick[0] = sfaf->tick[2] = 0;
            faser = 0;
    }   }
    if  (-32767 > faser) faser = -32767;
    if  (32767 < faser) faser = 32767;
    return(faser);
#   undef blin_internal_flags
}

static int
/*****************************************************************************************************/
dovol(auwith_vol *vol) {                                                                         /****
 *****************************************************************************************************/
    int v = 0;

    if  (0 > vol->min) {
        if  (v < -vol->min) v = -vol->min;
    } else {
        if  (v < vol->min) v = vol->min;
    }
    if  (0 > vol->max) {
        if  (v < -vol->max) v = -vol->max;
    } else {
        if  (v < vol->max) v = vol->max;
    }
    return(v);
}

void
/*****************************************************************************************************
 *****************************************************************************************************/
auwith_matrix(auwith_pid *pidm, FILE *fo) {                                                      /****
 *****************************************************************************************************
 *****************************************************************************************************/
    auwith_vol *inpvol, *outvol;
    BLIN_flag f;
    u_int i, o;

    inpvol = (auwith_vol*)&pidm->matrix[pidm->nout * pidm->ninp];
    outvol = &inpvol[pidm->ninp];
    fprintf(fo, "        ");
    for (o = 0; o < pidm->nout; ++o) fprintf(fo, " %5u", dovol(&outvol[o]));
    fprintf(fo, "\n      i\\o");
    for (o = 0; o < pidm->nout; ++o) fprintf(fo, "%s%5u", o ? " " : "", o);
    fprintf(fo, "\n");
    for (i = 0; i < pidm->ninp; ++i) {
        fprintf(fo, "%5u %2u", dovol(&inpvol[i]), i);
        for (o = 0; o < pidm->nout; ++o) {
            f = auwith_cella(pidm, i, o)->flags;
            if  ((auwith_cella(pidm, i, o)->inp != i) || (auwith_cella(pidm, i, o)->out != o)) {
                fprintf(fo, "<><><>");
            } else if (f & AUWITH_USED) {
                if  (f & AUWITH_ALGO) {
                    fprintf(fo, "%6u", f & AUWITH_LEVEL);
                } else {
                    fprintf(fo, "    ^^");
                }
            } else if (f & AUWITH_ALGO) {
                fprintf(fo, "    vv");
            } else {
                fprintf(fo, "    --");
        }   }
        fprintf(fo, "\n");
    }
    fflush(fo);
}

void
/*****************************************************************************************************
 *****************************************************************************************************/
auwith_Tinit(auwith_opd *opd) {                                                                  /****
 *****************************************************************************************************
 *****************************************************************************************************/
    if  (opd->T) {
        int e = errno;
        struct {
            char        t[8];
            auwith_time time;
        } mark = {"\020WITH T\000", 0};

        if  (0 > (opd->t = open(opd->T, O_WRONLY | O_APPEND | O_CREAT | O_TRUNC | O_EXLOCK, 0666))) {
            ifBLIN_QW0("open T=%s~", opd->T);
        }
        mark.time = auwith_future(0);
        if  (16 != write(opd->t, &mark, 16)) {
            ifBLIN_QW0("write mark T");
        }
        errno = e;
}   }

void
/*****************************************************************************************************
 *****************************************************************************************************/
auwith_Tmark(auwith_opd *opd, int32_t c) {                                                       /****
 *****************************************************************************************************
 *****************************************************************************************************/
    if  (opd->T) {
        int e = errno;
        struct {
            char        t[4];
            int32_t     c;
            auwith_time time;
        } mark = {"\0\0\0\015", 0, 0};

        mark.time = auwith_future(0);
        mark.c = c;
        if  (13 != write(opd->t, ((char*)&mark) + 3, 13)) {
            ifBLIN_QW0("write T");
        }
        errno = e;
}   }
