/*-
 * Copyright (C)2014 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2014 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n\
@(#)V.M "VERS;
static const char rcsid[] = "$Id: file.c,v 1.1 2014/08/14 09:34:37 babolo Exp $";
#endif /* not lint */

#define BLIN_COMPAT  3
#define Bpars_COMPAT 3

#include <sys/types.h>
#include <sys/event.h>
#include <sys/time.h>
#include <sysexits.h>
#include <strings.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include "auwith_int.h"

#define AUWITH_CONT AUWITH_LOCAL

                        /*****************************/
typedef struct {        /********* auwith_ent_file ***/
    auwith_ent      common;
    const char     *name;
    int             fd;
    auwith_buf     *buf[AUWITH_MACH];
    void           *b;
} auwith_ent_file;
                        /*****************************/
typedef struct {        /********* auwith_io_file ****/
    auwith_io       common;
    u_int16_t       chan;
} auwith_io_file;

static void
/*****************************************************************************************************/
usage(void) {                                                                                    /****
 *****************************************************************************************************/
    fprintf( stderr
           , "auwithd file V%u @BABOLO " DATE "\n"
             "Usage: auwithd [-options] file [-options]\n"
             "   -A              A law\n"
             "   -b num          sample width in bits (default %u)\n"
             "   -c num          channels number\n"
             "   -f file         file name\n"
             "   -E -X XX        use XX as env name for opt's arg\n"
             "   -i in           matrix's input number\n"
             "   -l              don't switch off at file end\n"
             "   -L              local phase count\n"
             "   -o out          matrix's output number\n"
             "   -q              silent\n"
             "   -r freq         samples rate in Hz (default %u)\n"
             "   -s num          block size in samples (default %u)\n"
             "   -u              u law\n"
             "   -v[v[v...]]     verbose\n"
             "   -w 0xXX         swap bits (0x01), doubles (0x02), nibbles (0x04) so on\n"
             "   -z TEXT         parce TEXT as arguments continuation\n"
             "   -Z TEXT         parce TEXT as arguments continuation with env substitution\n"
             "   -= mode         internal diagnostic\n"
             "   -?              this text\n"
           , auwith_def_file.mark.ver
           , AUWITH_BITS
           , AUWITH_RATE
           , AUWITH_BLOCK
           )
    ;
}

static int
/*****************************************************************************************************/
parsin(auwith_opd *opd) {                                                                        /****
 *****************************************************************************************************/
    struct kevent ev = {0, EVFILT_VNODE, EV_ADD, NOTE_DELETE | NOTE_RENAME, 0, NULL};
    char *flar = "Ab:c:f:i:lLo:qr:s:uvw:=:?";
    int blsize, samplewidth;
    u_int16_t inpp, oupp;
    auwith_ent_file *ent;
    int c, ex = EX_OK;

    ifBLIN_QX3("+");
    if  (!opd) {
        ifBLIN_QX0("no opd");
        ERROUT(EX_DATAERR, EFAULT);
    }
#   define blin_internal_flags (opd->flags & BLIN_MASK)
    if  (!opd->bos) {
        ifBLIN_QX0("no bos");
        ERROUT(EX_DATAERR, EFAULT);
    }
    if  (!(ent = calloc(1, sizeof(auwith_ent_file)))) {
        ifBLIN_QX0("Malloc failed #1");
        ERROUT(EX_OSERR, ENOMEM);
    }
    ent->common.flags = (opd->flags & ~AUWITH_CAP_MASK) | (auwith_def_file.flags & AUWITH_CAP_MASK);
    ent->common.plug = &auwith_def_file;
    ent->common.block = AUWITH_BLOCK;
    ent->common.bits = AUWITH_BITS;
    ent->common.rate = AUWITH_RATE;
    if  (!!babolo_chopts(opd->bos, 0, flar)) {
        ifBLIN_QW0("chopts");
        EXOUT(EX_SOFTWARE);
    }
    ex = EX_OK;
    if  (opd->flags & AUWITH_DUMPOP) babolo_dumpopts(opd->bos);
    while (!!(c = babolo_getopts(opd->bos))) {
        switch (c) {
        case 'A': ent->common.flags |= AUWITH_CAP_ALAW;
                ; break;
        case 'b': ent->common.bits = babolo_getoptunum(opd->bos, 0);
                ; break;
        case 'c': if  (!!ent->common.nchan) {
                      ifBLIN_QX0("redefine nchan");
                      ERROUT(EX_USAGE, ENOTEMPTY);
                  }
                ; ent->common.nchan = babolo_getoptunum(opd->bos, 0);
                ; inpp = 0;
                ; oupp = 0;
                ; break;
        case 'f': ent->name = babolo_getoptsarg(opd->bos);
                ; break;
        case 'i': if  (!(ent->common.flags & AUWITH_CAP_INP)) {
                      ifBLIN_QX0("file can be in or out, not both");
                      ERROUT(EX_USAGE, EADDRINUSE);
                  }
                ; ent->common.flags &= ~AUWITH_CAP_OUT;
                ; if  (inpp >= 1) {
                      ifBLIN_QX0("chan %u >= %u", inpp, 1);
                      EXOUT(EX_USAGE);
                  }
                ; if  (!(ent->common.in[inpp] = calloc(1, sizeof(auwith_io_file)))) {
                      ifBLIN_QX0("Malloc failed #2");
                      ERROUT(EX_OSERR, ENOMEM);
                  }
                ; ent->common.in[inpp]->flags = ent->common.flags;
                ; ent->common.in[inpp]->ent = (auwith_ent*)ent;
                ; ent->common.in[inpp]->sock = babolo_getoptunum(opd->bos, 0);
                ; if  (ent->common.in[inpp]->sock >= opd->inpachsz) {
                      ifBLIN_QX0( "-i[%u] must be < %u, now %u"
                                , inpp, opd->inpachsz, ent->common.in[inpp]->sock
                                );
                      ERROUT(EX_USAGE, EINVAL);
                  }
                ; ++inpp;
                ; break;
        case 'l': ent->common.flags |= AUWITH_CONT;
                ; break;
        case 'L': ent->common.flags ^= AUWITH_CFAF;
                ; break;
        case 'o': if  (!(ent->common.flags & AUWITH_CAP_OUT)) {
                      ifBLIN_QX0("file can be in or out, not both");
                      ERROUT(EX_USAGE, EADDRINUSE);
                  }
                ; ent->common.flags &= ~AUWITH_CAP_INP;
                ; if  (oupp >= ent->common.nchan) {
                      ifBLIN_QX0("chan %u >= %u", oupp, ent->common.nchan);
                      EXOUT(EX_USAGE);
                  }
                ; if  (!(ent->common.ou[oupp] = calloc(1, sizeof(auwith_io_file)))) {
                      ifBLIN_QX0("Malloc failed #3");
                      ERROUT(EX_OSERR, ENOMEM);
                  }
                ; ent->common.ou[oupp]->flags = ent->common.flags;
                ; ent->common.ou[oupp]->ent = (auwith_ent*)ent;
                ; ((auwith_io_file*)ent->common.ou[oupp])->chan = oupp;
                ; ent->common.ou[oupp]->sock = babolo_getoptunum(opd->bos, 0);
                ; if  (ent->common.ou[oupp]->sock >= opd->outachsz) {
                      ifBLIN_QX0( "-o[%u] must be < %u, now %u"
                                , oupp, opd->outachsz, ent->common.ou[oupp]->sock
                                );
                      ERROUT(EX_USAGE, EINVAL);
                  }
                ; ++oupp;
                ; break;
        case 'q': BLIN_QUIET(ent->common.flags);
                ; break;
        case 'r': ent->common.rate = babolo_getoptunum(opd->bos, 0);
                ; break;
        case 's': ent->common.block = babolo_getoptunum(opd->bos, 0);
                ; break;
        case 'u': ent->common.flags |= AUWITH_CAP_uLAW;
                ; break;
        case 'v': BLIN_VERBOSE(ent->common.flags);
                ; break;
        case 'w': ent->common.flags &= ~AUWITH_CAP_SWAP;
                ; ent->common.flags |= (babolo_getoptunum(opd->bos, 0) << 6) & AUWITH_CAP_SWAP;
                ; break;
        case '=': switch(babolo_getoptsarg(opd->bos)[0]) {
                  case '0': blin_ctl(BLIN_CTL_DUMP);           exit(EX_OK);
                  case '1': babolo_optest(flar);               exit(EX_OK);
                  case '2': babolo_optext(opd->bos);           exit(EX_OK);
                  case '3': babolo_dumpopts(opd->bos);         exit(EX_OK);
                  }
                ; break;
        case '?': usage();
                ; break;
        default : ifBLIN_QX0("Illegal flag -%c", opd->bos->c);
                ; usage();
                ; ERROUT(EX_USAGE, EINVAL);
    }   }
    if  (!ent->common.bits || (ent->common.bits > 16)) {
        ifBLIN_QX0("-b must be >0 and <=16, now %u", ent->common.bits);
        ERROUT(EX_USAGE, EINVAL);
    }
    if  (!ent->common.rate || (ent->common.rate > 48000)) {
        ifBLIN_QX0("-r must be >0 and <=48000, now %u", ent->common.rate);
        ERROUT(EX_USAGE, EINVAL);
    }
    for (inpp = 0; !ex && (inpp < ent->common.nchan); ++inpp) {
        if  (!!(ex = auwith_plumb(opd, (auwith_io*)ent->common.in[inpp], 0))) {
            ifBLIN_QW0("auwith_plumbi");
            goto out;
    }   }
    for (oupp = 0; !ex && (oupp < ent->common.nchan); ++oupp) {
        if  (!!(ex = auwith_plumb(opd, (auwith_io*)ent->common.ou[oupp], 1))) {
            ifBLIN_QW0("auwith_plumbo");
            goto out;
    }   }
    if  (!(ent->common.flags & AUWITH_CAP_OUT) && (ent->common.flags & AUWITH_CAP_INP)) {
        if  (0 > (ent->fd = open(ent->name, O_RDONLY))) {
            ifBLIN_QW0("read open %s", ent->name);
            EXOUT(EX_IOERR);
        }
    } else if ((ent->common.flags & AUWITH_CAP_OUT) && !(ent->common.flags & AUWITH_CAP_INP)) {
        if  (0 > (ent->fd = open(ent->name, O_WRONLY | O_APPEND | O_CREAT, 0666))) {
            ifBLIN_QW0("write open %s", ent->name);
            EXOUT(EX_IOERR);
    }   }
    samplewidth = (ent->common.flags & (AUWITH_CAP_ALAW | AUWITH_CAP_uLAW))
                ? 1
                : (ent->common.bits + 7) / 8
    ;
    blsize = samplewidth * ent->common.nchan * ent->common.block;
    if  (!(ent->b = malloc(blsize))) {
        ifBLIN_QX0("Malloc failed #4");
        ERROUT(EX_OSERR, ENOMEM);
    }
    ev.ident = ent->fd;
    if  (!!(ex = auwith_queue(opd, &ev, &ent->common, 0))) {
        ifBLIN_QW0("queue");
        goto out;
    }
out:
    if  (!!ex && !!ent) {
        for (inpp = 0; inpp < ent->common.nchan; ++inpp) {
            if  (!!ent->common.in[inpp]) free(ent->common.in[inpp]);
        }
        for (oupp = 0; oupp < ent->common.nchan; ++oupp) {
            if  (!!ent->common.ou[oupp]) free(ent->common.ou[oupp]);
        }
        free(ent);
    }
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

static void
/*****************************************************************************************************/
deconf(auwith_opd *opd, auwith_io *ioc, FILE *fo) {                                              /****
 *****************************************************************************************************/
    auwith_io_file *io = (auwith_io_file*)ioc;

#   define blin_internal_flags (ioc->flags & BLIN_MASK)
    fprintf( fo, "[%u] file %uHz, %usamples, chan %u, %s(%d)%s, sw=%02X^%08X\n"
           , io->common.sock
           , io->common.ent->rate
           , io->common.ent->block
           , io->chan
           , ((auwith_ent_file*)io->common.ent)->name
           , ((auwith_ent_file*)io->common.ent)->fd
           ,   (io->common.ent->flags & AUWITH_CAP_ALAW) ? ", G711A"
             : (io->common.ent->flags & AUWITH_CAP_uLAW) ? ", G711u" : ""
           , (io->common.ent->flags & AUWITH_CAP_SWAP) >> AUWITH_CAP_SFULLS
           , io->common.ent->xormask
           );
    fprintf(fo, "    ");
    auwith_capand(opd->flags, fo, 0);
    fprintf(fo, " ");
    auwith_capand(ioc->ent->flags, fo, 1);
    fprintf(fo, " ");
    auwith_capand(ioc->flags, fo, 1);
    fprintf(fo, "\n");
#   undef blin_internal_flags
}

static int
/*****************************************************************************************************/
activ(auwith_opd *opd, auwith_ent *ent, int kind) {                                              /****
 *****************************************************************************************************/
    int ex = EX_OK;

#   define blin_internal_flags (ent->flags & BLIN_MASK)
    ifBLIN_QX3("+ %d", kind);
    if  (kind != AUWITH_KIND_KQ) {
        ifBLIN_QX0("kind = %d", kind);
        ERROUT(EX_SOFTWARE, EINVAL);
    }
    auwith_Tmark(opd, 16);
    close(((auwith_ent_file*)ent)->fd);
    auwith_Tmark(opd, 17);
    if  (!(ent->flags & AUWITH_CAP_OUT) && (ent->flags & AUWITH_CAP_INP)) {
        if  (0 > (((auwith_ent_file*)ent)->fd = open(((auwith_ent_file*)ent)->name, O_RDONLY))) {
            ifBLIN_QW0("read open %s", ((auwith_ent_file*)ent)->name);
            EXOUT(EX_IOERR);
        }
    } else if ((ent->flags & AUWITH_CAP_OUT) && !(ent->flags & AUWITH_CAP_INP)) {
        if  (0 > (((auwith_ent_file*)ent)->fd = open(((auwith_ent_file*)ent)->name, O_WRONLY | O_APPEND | O_CREAT, 0666))) {
            ifBLIN_QW0("write open %s", ((auwith_ent_file*)ent)->name);
            EXOUT(EX_IOERR);
    }   }
    auwith_Tmark(opd, 18);
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

static int
/*****************************************************************************************************/
passiv(auwith_opd *opd, auwith_cell *cell, int kind) {                                           /****
 *****************************************************************************************************/
    int blsize, samplewidth, p, ex = EX_OK;
    auwith_ent_file *ent;
    auwith_buf *buf;

#   define blin_internal_flags (cell->flags & BLIN_MASK)
    ifBLIN_QX3("+ %d", kind);
    if  (kind != AUWITH_KIND_WRITE) goto out;
    ent = (auwith_ent_file*)opd->outach[cell->out]->ent;
    samplewidth = (ent->common.flags & (AUWITH_CAP_ALAW | AUWITH_CAP_uLAW))
                ? 1
                : (ent->common.bits + 7) / 8
    ;
    blsize = samplewidth * ent->common.nchan * ent->common.block;
    if  (!(buf = auwith_prepare(opd, &ent->common))) {
        ifBLIN_QW0("auwith_prepare");
        EXOUT(EX_OSERR);
    }
    p = 0;
    if  (!!buf->size) {
        auwith_Tmark(opd, 12);
        p = write(ent->fd, buf->samples, buf->size);
        auwith_Tmark(opd, 13);
        if  (0 > p) {
            ifBLIN_QW0("write");
            EXOUT(EX_IOERR);
        } else if (p != buf->size) {
            ifBLIN_QX1("write %d <> %d", p, buf->size);
    }   }
    auwith_relebuf(buf);
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

static auwith_buf *
/*****************************************************************************************************/
reed(auwith_opd *opd, auwith_cell *cell) {                                                       /****
 *****************************************************************************************************/
    int blsize, samplewidth, p, ex = EX_OK;
    auwith_ent_file *ent;
    auwith_hc *is;

#   define blin_internal_flags (cell->flags & BLIN_MASK)
    ifBLIN_QX3("+");
    is = auwith_cadd(opd, cell);
    ent = (auwith_ent_file*)opd->inpach[cell->inp]->ent;
    samplewidth = (ent->common.flags & (AUWITH_CAP_ALAW | AUWITH_CAP_uLAW))
                ? 1
                : (ent->common.bits + 7) / 8
    ;
    blsize = samplewidth * ent->common.nchan * ent->common.block;
    auwith_Tmark(opd, 14);
    p = pread(ent->fd, ent->b, blsize, is->offset);
    auwith_Tmark(opd, 15);
    if  (0 > p) {
        ifBLIN_QW0("pread");
        goto out;
    } else if (p > blsize) {
        ifBLIN_QX0("pread %d > %d", p, blsize);
        ERROUT(EX_IOERR, EIO);
    } else if (p < blsize) {
        ifBLIN_QX1("pread %d < %d", p, blsize);
    }
    is->offset += p;
    if  (!!p) {
    } else if (ent->common.flags & AUWITH_CONT) {
        bzero(ent->b, blsize);
        p = blsize;
    } else {
        cell->flags &= ~AUWITH_USED;
    }
    if  (!!(ex = auwith_postpare(opd, &ent->common, ent->b, p))) {
        ifBLIN_QW0("auwith_postpare");
        goto out;
    }
    errno = 0;
out:
    ifBLIN_QX3("-");
    return(NULL);
#   undef blin_internal_flags
}

/*****************************************************************************************************
 *****************************************************************************************************/
auwith_plug auwith_def_file =                                                                    /****
 *****************************************************************************************************
 *****************************************************************************************************/
{ "file"
, AUWITH_MARK
, parsin
, deconf
, activ
, passiv
, reed
, NULL
, AUWITH_CAP_INP | AUWITH_CAP_OUT
, 0
};
